import shutil
import tempfile

from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.select import Select


def cleanup_temp_dir(temp_dir: tempfile.TemporaryDirectory) -> None:
    """
    Deletes the temporary directory and all its contents.

    Args:
        temp_dir (tempfile.TemporaryDirectory): The temporary directory to delete.
    """
    if temp_dir:
        try:
            temp_dir.cleanup()
        except OSError:
            shutil.rmtree(temp_dir.name, ignore_errors=True)


def element_as_select(element: WebElement) -> Select:
    """Wraps a WebElement in a Select object.

    Args:
        element (WebElement): The element to wrap.

    Returns:
        Select: The Select object.
    """
    return Select(element)



from .bot import WebBot, Browser, BROWSER_CONFIGS, By, PageLoadStrategy  # noqa: F401, F403
from .parsers import table_to_dict, data_from_row, sanitize_header  # noqa: F401, F403
from .util import element_as_select  # noqa: F401, F403

from botcity.web._version import get_versions
__version__ = get_versions()['version']
del get_versions



import atexit
import json
import os
import tempfile
from typing import Dict

from selenium.webdriver import Chrome  # noqa: F401, F403
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

from ..util import cleanup_temp_dir


def default_options(headless=False, download_folder_path=None, user_data_dir=None,
                    page_load_strategy="normal") -> ChromeOptions:
    """Retrieve the default options for this browser curated by BotCity.

    Args:
        headless (bool, optional): Whether or not to use the headless mode. Defaults to False.
        download_folder_path (str, optional): The default path in which to save files.
            If None, the current directory is used. Defaults to None.
        user_data_dir ([type], optional): The directory to use as user profile.
            If None, a new temporary directory is used. Defaults to None.
        page_load_strategy (str, optional): The page load strategy. Defaults to "normal".

    Returns:
        ChromeOptions: The Chrome options.
    """
    chrome_options = ChromeOptions()
    try:
        page_load_strategy = page_load_strategy.value
    except AttributeError:
        page_load_strategy = page_load_strategy
    chrome_options.page_load_strategy = page_load_strategy
    chrome_options.add_argument("--remote-debugging-port=0")
    chrome_options.add_argument("--no-first-run")
    chrome_options.add_argument("--no-default-browser-check")
    chrome_options.add_argument("--disable-background-networking")
    chrome_options.add_argument("--disable-background-timer-throttling")
    chrome_options.add_argument("--disable-client-side-phishing-detection")
    chrome_options.add_argument("--disable-default-apps")
    chrome_options.add_argument("--disable-hang-monitor")
    chrome_options.add_argument("--disable-popup-blocking")
    chrome_options.add_argument("--disable-prompt-on-repost")
    chrome_options.add_argument("--disable-syncdisable-translate")
    chrome_options.add_argument("--metrics-recording-only")
    chrome_options.add_argument("--safebrowsing-disable-auto-update")

    # Disable What's New banner for new chrome installs
    chrome_options.add_argument("--disable-features=ChromeWhatsNewUI")

    chrome_options.add_argument("--disable-blink-features=AutomationControlled")

    # Disable banner for Browser being remote-controlled
    chrome_options.add_experimental_option("excludeSwitches", ["enable-automation"])
    chrome_options.add_experimental_option('useAutomationExtension', False)

    if headless:
        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--hide-scrollbars")
        chrome_options.add_argument("--mute-audio")

    # Check if user is root
    try:
        # This is only valid with Unix
        if os.geteuid() == 0:
            chrome_options.add_argument("--no-sandbox")
    except AttributeError:
        pass

    if not user_data_dir:
        temp_dir = tempfile.TemporaryDirectory(prefix="botcity_")
        user_data_dir = temp_dir.name
        atexit.register(cleanup_temp_dir, temp_dir)

    chrome_options.add_argument(f"--user-data-dir={user_data_dir}")

    if not download_folder_path:
        download_folder_path = os.getcwd()

    app_state = {
        'recentDestinations': [{
            'id': 'Save as PDF',
            'origin': 'local',
            'account': ''
        }],
        'selectedDestinationId': 'Save as PDF',
        'version': 2
    }

    # Set the Downloads default folder
    prefs = {
        "printing.print_preview_sticky_settings.appState": json.dumps(app_state),
        "download.default_directory": download_folder_path,
        "savefile.default_directory": download_folder_path,
        "printing.default_destination_selection_rules": {
            "kind": "local",
            "namePattern": "Save as PDF",
        },
        "safebrowsing.enabled": True,
        "credentials_enable_service": False,
        "profile.password_manager_enabled": False,
        "plugins.always_open_pdf_externally": True
    }

    chrome_options.add_experimental_option("prefs", prefs)
    chrome_options.add_argument(
        "user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36"
    )

    chrome_options.add_argument("--kiosk-printing")

    return chrome_options


def default_capabilities() -> Dict:
    """Fetch the default capabilities for this browser.

    Returns:
        Dict: Dictionary with the default capabilities defined.
    """
    return DesiredCapabilities.CHROME.copy()


def wait_for_downloads(driver):
    """Wait for all downloads to finish.
    *Important*: This method overwrites the current page with the downloads page.
    """
    if not driver.current_url.startswith("chrome://downloads"):
        driver.get("chrome://downloads/")
    return driver.execute_script("""
        var items = document.querySelector('downloads-manager')
            .shadowRoot.getElementById('downloadsList').items;
        if (items.every(e => e.state === "COMPLETE"))
            return items.map(e => e.fileUrl || e.file_url);
        """)



import collections
import string
from typing import Dict, List, Optional
from selenium.webdriver.remote.webelement import WebElement


def data_from_row(row: WebElement, cell_tag="td", cell_xpath=None) -> List[str]:
    """Extract data from a row and return it as a list.

    Args:
        row (WebElement): The row element.
        cell_tag (str, optional): The HTML tag associated with the row cells. Defaults to "td".
        cell_xpath (str, optional): The XPath expression associated with the row cels. Defaults to None.
            If informed, overwrites the `cell_tag` definition.

    Returns:
        list: List of strings with the contents.
    """
    if cell_xpath:
        return [
            col.text for col in row.find_elements_by_xpath(cell_xpath)
        ]

    return [
        col.text for col in row.find_elements_by_tag_name(cell_tag)
    ]


def sanitize_header(labels: List[str]):
    """Sanitize header labels."""
    # Handle Treat Empty Header
    for idx, label in enumerate(labels):
        if label.strip():
            # make it lowercase
            label = label.lower()

            # remove punctuations
            label = ''.join([l for l in label if l not in string.punctuation])  # noqa: E741

            # replace spaces with underscores
            label = label.replace(" ", "_")
        else:
            label = f"col_{idx}"
        labels[idx] = label

    # Deduplicate by adding _1, _2, _3 to repeated labels
    counts = {k: v for k, v in collections.Counter(labels).items() if v > 1}
    for i in reversed(range(len(labels))):
        item = labels[i]
        if item in counts and counts[item]:
            labels[i] = f"{item}_{counts[item]}"
            counts[item] -= 1

    return labels


def table_to_dict(table: WebElement, has_header: bool = True,
                  skip_rows: int = 0, header_tag: str = "th",
                  cell_xpath: Optional[str] = None) -> List[Dict]:
    """Convert a table WebElement to a dict of lists.

    Args:
        table (WebElement): The table element.
        has_header (bool, optional): Whether or not to parse a header. Defaults to True.
        skip_rows (int, optional): Number of rows to skip from the top. Defaults to 0.
        header_tag (str, optional): The HTML tag associated with the header cell. Defaults to "th".
        cell_xpath (str, optional): Optional cell XPath selector for complex row constructions.
            If `cell_xpath` is not informed, the row data will come from `<td>` elements.

    Returns:
        list: List with dict for each row.
    """

    # Collect all rows from table
    rows = table.find_elements_by_tag_name("tr")

    # Skip rows if informed
    if skip_rows:
        rows = rows[skip_rows:]

    if cell_xpath and not cell_xpath.startswith('.'):
        # Convert into relative xpath
        cell_xpath = f'.{cell_xpath}'

    # Parse header labels
    if has_header:
        # Read header labels
        labels = data_from_row(rows[0], cell_tag=header_tag)
        # Sanitize headers
        labels = sanitize_header(labels)
        # Skip the header
        rows = rows[1:]
    else:
        # Make up header labels
        if cell_xpath:
            cols = rows[0].find_elements_by_xpath(cell_xpath)
        else:
            cols = rows[0].find_elements_by_tag_name("td")

        num_cols = len(cols)
        labels = [f"col_{i}" for i in range(num_cols)]

    # Assemble output dictionary
    out_list = []
    for row in rows:
        row_data = data_from_row(row, cell_xpath=cell_xpath)
        out_list.append(dict(zip(labels, row_data)))

    return out_list



import base64
import functools
import glob
import io
import json
import logging
import os
import platform
import random
import re
import shutil
import time
from typing import List
from contextlib import contextmanager

from botcity.base import BaseBot, State
from botcity.base.utils import only_if_element
from bs4 import BeautifulSoup
from PIL import Image
from selenium.common.exceptions import InvalidSessionIdException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.wait import WebDriverWait, TimeoutException, NoSuchElementException
from selenium.webdriver.support import expected_conditions as EC

from . import config, cv2find
from .browsers import BROWSER_CONFIGS, Browser, PageLoadStrategy

try:
    from botcity.maestro import BotMaestroSDK
    MAESTRO_AVAILABLE = True
except ImportError:
    MAESTRO_AVAILABLE = False


logger = logging.getLogger(__name__)


class WebBot(BaseBot):
    KEYS = Keys
    DEFAULT_DIMENSIONS = (1600, 900)
    """
    Base class for Web Bots.
    Users must implement the `action` method in their classes.

    Attributes:
        state (State): The internal state of this bot.
        maestro (BotMaestroSDK): an instance to interact with the BotMaestro server.

    """

    def __init__(self, headless=False):
        self.state = State()
        self.maestro = BotMaestroSDK() if MAESTRO_AVAILABLE else None

        self._browser = Browser.CHROME
        self._options = None
        self._capabilities = None
        self._driver_path = None

        self._driver = None
        self._headless = headless
        self._page_load_strategy = PageLoadStrategy.NORMAL

        self._clipboard = ""

        # Stub mouse coordinates
        self._html_elem = None
        self._x = 0
        self._y = 0

        # State for Key modifiers
        self._shift_hold = False

        self._download_folder_path = os.getcwd()

    @property
    def driver(self):
        """
        The WebDriver driver instance.

        Returns:
            driver (WebDriver): The WebDriver driver instance.

        """
        return self._driver

    @property
    def driver_path(self):
        return self._driver_path

    @driver_path.setter
    def driver_path(self, driver_path):
        """
        The webdriver executable path.

        Args:
            driver_path (str): The full path to the proper webdriver path used for the selected browser.
                If set to None, the code will look into the PATH for the proper file when starting the browser.
        """
        driver_path = os.path.abspath(os.path.expanduser(os.path.expandvars(driver_path)))
        if driver_path and not os.path.isfile(driver_path):
            raise ValueError("Invalid driver_path. The file does not exist.")
        self._driver_path = driver_path

    @property
    def browser(self):
        """
        The web browser to be used.

        Returns:
            browser (Browser): The web browser to be used.
        """
        return self._browser

    @browser.setter
    def browser(self, browser):
        """
        The web browser to be used.

        Args:
            browser (Browser): The name of web browser to be used from the Browser enum.
        """
        self._browser = browser

    @property
    def options(self):
        """
        The options to be passed down to the WebDriver when starting the browser.

        Returns:
            options (Options): The browser specific options to be used.
        """
        return self._options

    @options.setter
    def options(self, options):
        """
        The options to be passed down to the WebDriver when starting the browser.

        Args:
            options (Options): The browser specific options to be used.
        """
        self._options = options

    @property
    def capabilities(self):
        """
        The capabilities to be passed down to the WebDriver when starting the browser.

        Returns:
            capabilities (Dict): The browser specific capabilities to be used.
        """
        return self._capabilities

    @capabilities.setter
    def capabilities(self, capabilities):
        """
        The capabilities to be passed down to the WebDriver when starting the browser.

        Args:
            capabilities (Dict): The browser specific capabilities to be used.
        """
        self._capabilities = capabilities

    @property
    def download_folder_path(self):
        return self._download_folder_path

    @download_folder_path.setter
    def download_folder_path(self, folder_path):
        """
        The download folder path to be used. Set it up before starting the Browser or browsing a URL or restart the
        browser after changing it.

        Args:
            folder_path (str): The desired download folder path.
        """
        self._download_folder_path = folder_path

    @property
    def headless(self):
        """
        Controls whether or not the bot will run headless.

        Returns:
            headless (bool): Whether or not to run the browser on headless mode.
        """
        return self._headless

    @headless.setter
    def headless(self, headless):
        """
        Controls whether or not the bot will run headless.

        Args:
            headless (boolean): If set to True will make the bot run headless.
        """
        if self._driver:
            logger.warning("Browser is running. Invoke stop_browser and start browser for changes to take effect.")
        self._headless = headless

    @property
    def page_load_strategy(self) -> PageLoadStrategy:
        """
        The page load strategy to be used.

        Returns:
            page_load_strategy (PageLoadStrategy): The page load strategy to be used.
        """
        return self._page_load_strategy

    @page_load_strategy.setter
    def page_load_strategy(self, page_load_strategy: PageLoadStrategy):
        """
        The page load strategy to be used.

        Args:
            page_load_strategy (PageLoadStrategy): The page load strategy to be used.
        """
        if self._driver:
            logger.warning("Browser is running. Invoke stop_browser and start browser for changes to take effect.")
        self._page_load_strategy = page_load_strategy

    def start_browser(self):
        """
        Starts the selected browser.
        """

        def check_driver():
            # Look for driver
            driver_name = BROWSER_CONFIGS.get(self.browser).get("driver")
            location = shutil.which(driver_name)
            if not location:
                raise RuntimeError(
                    f"{driver_name} was not found. Please make sure to have it on your PATH or set driver_path")
            return location

        # Specific webdriver class for a given browser
        driver_class = BROWSER_CONFIGS.get(self.browser).get("class")
        # Specific default options method for a given browser
        func_def_options = BROWSER_CONFIGS.get(self.browser).get("options")
        # Specific capabilities method for a given browser
        func_def_capabilities = BROWSER_CONFIGS.get(self.browser).get("capabilities")

        opt = self.options or func_def_options(
            self.headless, self._download_folder_path, None, self.page_load_strategy
        )
        cap = self.capabilities or func_def_capabilities()
        self.options = opt
        self.capabilities = cap
        driver_path = self.driver_path or check_driver()
        self.driver_path = driver_path

        self._driver = driver_class(options=opt, desired_capabilities=cap, executable_path=driver_path)
        self.set_screen_resolution()

    def stop_browser(self):
        """
        Stops the Chrome browser and clean up the User Data Directory.

        Warning:
            After invoking this method, you will need to reassign your custom options and capabilities.
        """
        if not self._driver:
            return
        self._driver.close()
        self._driver.quit()
        self.options = None
        self.capabilities = None
        self._driver = None

    def set_screen_resolution(self, width=None, height=None):
        """
        Configures the browser dimensions.

        Args:
            width (int): The desired width.
            height (int): The desired height.
        """
        dimensions = (width or self.DEFAULT_DIMENSIONS[0], height or self.DEFAULT_DIMENSIONS[1])

        if self.headless:
            # When running headless the window size is the viewport size
            window_size = dimensions
        else:
            # When running non-headless we need to account for the borders and etc
            # So the size must be bigger to have the same viewport size as before
            window_size = self._driver.execute_script("""
                return [window.outerWidth - window.innerWidth + arguments[0],
                  window.outerHeight - window.innerHeight + arguments[1]];
                """, *dimensions)
        self._driver.set_window_size(*window_size)

    def _webdriver_command(self, command, params=None, req_type="POST"):
        """
        Execute a webdriver command.

        Args:
            command (str): The command URL after the session part
            params (dict): The payload to be serialized and sent to the webdriver. Defaults to None.
            req_type (str, optional): The type of request to be made. Defaults to "POST".

        Returns:
            str: The value of the response
        """
        if not params:
            params = {}

        resource = f"/session/{self.driver.session_id}/{command}"
        url = self.driver.command_executor._url + resource
        body = json.dumps(params)
        response = self.driver.command_executor._request(req_type, url, body)

        if not response:
            raise Exception(response.get('value'))

        return response.get('value')

    ##########
    # Display
    ##########

    def get_screen_image(self, region=None):
        """
        Capture and returns a screenshot from the browser.

        Args:
            region (tuple): A tuple containing the left, top, width and height
                to crop the screen image.

        Returns:
            image (Image): The screenshot Image object.
        """
        if not region:
            region = (0, 0, 0, 0)

        x = region[0]
        y = region[1]
        width = region[2] or self._get_page_size()[0]
        height = region[3] or self._get_page_size()[1]

        try:
            data = self._driver.get_screenshot_as_base64()
            image_data = base64.b64decode(data)
            img = Image.open(io.BytesIO(image_data))
        except:  # noqa: E722
            img = Image.new("RGB", (width, height))

        img = img.crop((x, y, x + width, y + height))
        return img

    def get_viewport_size(self):
        """
        Returns the browser current viewport size.

        Returns:
            width (int): The current viewport width.
            height (int): The current viewport height.
        """
        # Access each dimension individually
        width = self._driver.get_window_size().get("width")
        height = self._driver.get_window_size().get("height")
        return width, height

    def _get_page_size(self):
        """
        Returns the browser current page size.

        Returns:
            width (int): The current page width.
            height (int): The current page height.
        """
        if not self._driver:
            return self.DEFAULT_DIMENSIONS

        width = self.execute_javascript("return window.innerWidth")
        height = self.execute_javascript("return window.innerHeight")
        return width, height

    def add_image(self, label, path):
        """
        Add an image into the state image map.

        Args:
            label (str): The image identifier
            path (str): The path for the image on disk
        """
        self.state.map_images[label] = path

    def get_image_from_map(self, label):
        """
        Return an image from teh state image map.

        Args:
            label (str): The image identifier

        Returns:
            Image: The Image object
        """
        path = self.state.map_images.get(label)
        if not path:
            raise KeyError('Invalid label for image map.')
        img = Image.open(path)
        return img

    def find_multiple(self, labels, x=None, y=None, width=None, height=None, *,
                      threshold=None, matching=0.9, waiting_time=10000, best=True, grayscale=False):
        """
        Find multiple elements defined by label on screen until a timeout happens.

        Args:
            labels (list): A list of image identifiers
            x (int, optional): Search region start position x. Defaults to 0.
            y (int, optional): Search region start position y. Defaults to 0.
            width (int, optional): Search region width. Defaults to screen width.
            height (int, optional): Search region height. Defaults to screen height.
            threshold (int, optional): The threshold to be applied when doing grayscale search.
                Defaults to None.
            matching (float, optional): The matching index ranging from 0 to 1.
                Defaults to 0.9.
            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
                Defaults to 10000ms (10s).
            best (bool, optional): Whether or not to keep looking until the best matching is found.
                Defaults to True.
            grayscale (bool, optional): Whether or not to convert to grayscale before searching.
                Defaults to False.

        Returns:
            results (dict): A dictionary in which the key is the label and value are the element coordinates in a
               NamedTuple.
        """

        def _to_dict(lbs, elems):
            return {k: v for k, v in zip(lbs, elems)}

        screen_w, screen_h = self._get_page_size()
        x = x or 0
        y = y or 0
        w = width or screen_w
        h = height or screen_h

        region = (x, y, w, h)

        results = [None] * len(labels)
        paths = [self._search_image_file(la) for la in labels]
        paths = [self._image_path_as_image(la) for la in paths]

        if threshold:
            # TODO: Figure out how we should do threshold
            print('Threshold not yet supported')

        if not best:
            # TODO: Implement best=False.
            print('Warning: Ignoring best=False for now. It will be supported in the future.')

        start_time = time.time()

        while True:
            elapsed_time = (time.time() - start_time) * 1000
            if elapsed_time > waiting_time:
                return _to_dict(labels, results)

            haystack = self.screenshot()
            helper = functools.partial(self._find_multiple_helper, haystack, region, matching, grayscale)

            results = [helper(p) for p in paths]

            results = [r for r in results]
            if None in results:
                continue
            else:
                return _to_dict(labels, results)

    def _find_multiple_helper(self, haystack, region, confidence, grayscale, needle):
        ele = cv2find.locate_all_opencv(
            needle, haystack, region=region, confidence=confidence, grayscale=grayscale
        )
        try:
            ele = next(ele)
        except StopIteration:
            ele = None
        return ele

    def find(self, label, x=None, y=None, width=None, height=None, *,
             threshold=None, matching=0.9, waiting_time=10000, best=True, grayscale=False):
        """
        Find an element defined by label on screen until a timeout happens.

        Args:
            label (str): The image identifier
            x (int, optional): Search region start position x. Defaults to 0.
            y (int, optional): Search region start position y. Defaults to 0.
            width (int, optional): Search region width. Defaults to screen width.
            height (int, optional): Search region height. Defaults to screen height.
            threshold (int, optional): The threshold to be applied when doing grayscale search.
                Defaults to None.
            matching (float, optional): The matching index ranging from 0 to 1.
                Defaults to 0.9.
            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
                Defaults to 10000ms (10s).
            best (bool, optional): Whether or not to keep looking until the best matching is found.
                Defaults to True.
            grayscale (bool, optional): Whether or not to convert to grayscale before searching.
                Defaults to False.

        Returns:
            element (NamedTuple): The element coordinates. None if not found.
        """
        return self.find_until(label=label, x=x, y=y, width=width, height=height, threshold=threshold,
                               matching=matching, waiting_time=waiting_time, best=best, grayscale=grayscale)

    def find_until(self, label, x=None, y=None, width=None, height=None, *,
                   threshold=None, matching=0.9, waiting_time=10000, best=True, grayscale=False):
        """
        Find an element defined by label on screen until a timeout happens.

        Args:
            label (str): The image identifier
            x (int, optional): Search region start position x. Defaults to 0.
            y (int, optional): Search region start position y. Defaults to 0.
            width (int, optional): Search region width. Defaults to screen width.
            height (int, optional): Search region height. Defaults to screen height.
            threshold (int, optional): The threshold to be applied when doing grayscale search.
                Defaults to None.
            matching (float, optional): The matching index ranging from 0 to 1.
                Defaults to 0.9.
            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
                Defaults to 10000ms (10s).
            best (bool, optional): Whether or not to keep looking until the best matching is found.
                Defaults to True.
            grayscale (bool, optional): Whether or not to convert to grayscale before searching.
                Defaults to False.

        Returns:
            element (NamedTuple): The element coordinates. None if not found.
        """
        self.state.element = None
        screen_w, screen_h = self._get_page_size()
        x = x or 0
        y = y or 0
        w = width or screen_w
        h = height or screen_h

        region = (x, y, w, h)

        element_path = self._search_image_file(label)
        element_path = self._image_path_as_image(element_path)

        if threshold:
            # TODO: Figure out how we should do threshold
            print('Threshold not yet supported')

        if not best:
            # TODO: Implement best=False.
            print('Warning: Ignoring best=False for now. It will be supported in the future.')

        start_time = time.time()

        while True:
            elapsed_time = (time.time() - start_time) * 1000
            if elapsed_time > waiting_time:
                return None
            haystack = self.get_screen_image()
            it = cv2find.locate_all_opencv(element_path, haystack_image=haystack,
                                           region=region, confidence=matching, grayscale=grayscale)
            try:
                ele = next(it)
            except StopIteration:
                ele = None

            if ele is not None:
                self.state.element = ele
                return ele

    def set_current_element(self, element: cv2find.Box):
        """
        Changes the current screen element the bot will interact when using click(), move(), and similar methods.

        This method is equivalent to self.state.element = element.

        Args:
            element (Box): A screen element from self.state.element or the find_all(as_list=True) method.
        """
        self.state.element = element

    def find_all(self, label, x=None, y=None, width=None, height=None, *,
                 threshold=None, matching=0.9, waiting_time=10000, grayscale=False, as_list: bool = False):
        """
        Find all elements defined by label on screen until a timeout happens.

        Args:
            label (str): The image identifier
            x (int, optional): Search region start position x. Defaults to 0.
            y (int, optional): Search region start position y. Defaults to 0.
            width (int, optional): Search region width. Defaults to screen width.
            height (int, optional): Search region height. Defaults to screen height.
            threshold (int, optional): The threshold to be applied when doing grayscale search.
                Defaults to None.
            matching (float, optional): The matching index ranging from 0 to 1.
                Defaults to 0.9.
            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
                Defaults to 10000ms (10s).
            grayscale (bool, optional): Whether or not to convert to grayscale before searching.
                Defaults to False.
            as_list (bool, Optional): If True, returns a list of element coordinates instead of a generator.
                Use set_active_element() to be able to interact with the found elements.
                This parameter must be True if you intend to run multiple find_all() concurrently.
                Defaults to False.

        Returns:
            elements (collections.Iterable[NamedTuple]): A generator with all element coordinates found.
                None if not found.
        """
        def deduplicate(elems):
            def find_same(item, items):
                x_start = item.left
                x_end = item.left + item.width
                y_start = item.top
                y_end = item.top + item.height
                similars = []
                for itm in items:
                    if itm == item:
                        continue
                    if (itm.left >= x_start and itm.left < x_end)\
                            and (itm.top >= y_start and itm.top < y_end):
                        similars.append(itm)
                        continue
                return similars

            index = 0
            while True:
                try:
                    dups = find_same(elems[index], elems[index:])
                    for d in dups:
                        elems.remove(d)
                    index += 1
                except IndexError:
                    break
            return elems

        self.state.element = None
        screen_w, screen_h = self._get_page_size()
        x = x or 0
        y = y or 0
        w = width or screen_w
        h = height or screen_h

        region = (x, y, w, h)

        element_path = self._search_image_file(label)
        element_path = self._image_path_as_image(element_path)

        if threshold:
            # TODO: Figure out how we should do threshold
            print('Threshold not yet supported')

        start_time = time.time()

        while True:
            elapsed_time = (time.time() - start_time) * 1000
            if elapsed_time > waiting_time:
                return None

            haystack = self.get_screen_image()
            it = cv2find.locate_all_opencv(element_path, haystack_image=haystack,
                                           region=region, confidence=matching, grayscale=grayscale)

            eles = [ele for ele in it]
            if not eles:
                continue
            eles = deduplicate(list(eles))

            # As List
            if as_list:
                return eles

            # As Generator
            for ele in eles:
                if ele is not None:
                    self.state.element = ele
                    yield ele
            break

    def find_text(self, label, x=None, y=None, width=None, height=None, *, threshold=None, matching=0.9,
                  waiting_time=10000, best=True):
        """
        Find an element defined by label on screen until a timeout happens.

        Args:
            label (str): The image identifier
            x (int, optional): Search region start position x. Defaults to 0.
            y (int, optional): Search region start position y. Defaults to 0.
            width (int, optional): Search region width. Defaults to screen width.
            height (int, optional): Search region height. Defaults to screen height.
            threshold (int, optional): The threshold to be applied when doing grayscale search.
                Defaults to None.
            matching (float, optional): The matching index ranging from 0 to 1.
                Defaults to 0.9.
            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
                Defaults to 10000ms (10s).
            best (bool, optional): Whether or not to keep looking until the best matching is found.
                Defaults to True.

        Returns:
            element (NamedTuple): The element coordinates. None if not found.
        """
        return self.find_until(label, x, y, width, height, threshold=threshold, matching=matching,
                               waiting_time=waiting_time, best=best, grayscale=True)

    def get_last_element(self):
        """
        Return the last element found.

        Returns:
            element (NamedTuple): The element coordinates (left, top, width, height)
        """
        return self.state.element

    def display_size(self):
        """
        Returns the display size in pixels.

        Returns:
            size (Tuple): The screen dimension (width and height) in pixels.
        """
        return self._get_page_size()

    def screenshot(self, filepath=None, region=None):
        """
        Capture a screenshot.

        Args:
            filepath (str, optional): The filepath in which to save the screenshot. Defaults to None.
            region (tuple, optional): Bounding box containing left, top, width and height to crop screenshot.

        Returns:
            Image: The screenshot Image object
        """
        img = self.get_screen_image(region)
        if filepath:
            img.save(filepath)
        return img

    def get_screenshot(self, filepath=None, region=None):
        """
        Capture a screenshot.

        Args:
            filepath (str, optional): The filepath in which to save the screenshot. Defaults to None.
            region (tuple, optional): Bounding box containing left, top, width and height to crop screenshot.

        Returns:
            Image: The screenshot Image object
        """
        return self.screenshot(filepath, region)

    def screen_cut(self, x, y, width=None, height=None):
        """
        Capture a screenshot from a region of the screen.

        Args:
            x (int): region start position x
            y (int): region start position y
            width (int): region width
            height (int): region height

        Returns:
            Image: The screenshot Image object
        """
        screen_size = self._get_page_size()
        x = x or 0
        y = y or 0
        width = width or screen_size[0]
        height = height or screen_size[1]
        img = self.screenshot(region=(x, y, width, height))
        return img

    def save_screenshot(self, path):
        """
        Saves a screenshot in a given path.

        Args:
            path (str): The filepath in which to save the screenshot

        """
        self.screenshot(path)

    def get_element_coords(self, label, x=None, y=None, width=None, height=None, matching=0.9, best=True):
        """
        Find an element defined by label on screen and returns its coordinates.

        Args:
            label (str): The image identifier
            x (int, optional): X (Left) coordinate of the search area.
            y (int, optional): Y (Top) coordinate of the search area.
            width (int, optional): Width of the search area.
            height (int, optional): Height of the search area.
            matching (float, optional): Minimum score to consider a match in the element image recognition process.
                Defaults to 0.9.
            best (bool, optional): Whether or not to search for the best value. If False the method returns on
                the first find. Defaults to True.

        Returns:
            coords (Tuple): A tuple containing the x and y coordinates for the element.
        """
        self.state.element = None
        screen_size = self._get_page_size()
        x = x or 0
        y = y or 0
        width = width or screen_size[0]
        height = height or screen_size[1]
        region = (x, y, width, height)

        if not best:
            print('Warning: Ignoring best=False for now. It will be supported in the future.')

        element_path = self._search_image_file(label)
        element_path = self._image_path_as_image(element_path)
        haystack = self.get_screen_image()

        it = cv2find.locate_all_opencv(element_path, haystack_image=haystack,
                                       region=region, confidence=matching)
        try:
            ele = next(it)
        except StopIteration:
            ele = None
        self.state.element = ele

        if ele:
            return ele.left, ele.top
        else:
            return None, None

    def get_element_coords_centered(self, label, x=None, y=None, width=None, height=None,
                                    matching=0.9, best=True):
        """
        Find an element defined by label on screen and returns its centered coordinates.

        Args:
            label (str): The image identifier
            x (int, optional): X (Left) coordinate of the search area.
            y (int, optional): Y (Top) coordinate of the search area.
            width (int, optional): Width of the search area.
            height (int, optional): Height of the search area.
            matching (float, optional): Minimum score to consider a match in the element image recognition process.
                Defaults to 0.9.
            best (bool, optional): Whether or not to search for the best value. If False the method returns on
                the first find. Defaults to True.

        Returns:
            coords (Tuple): A tuple containing the x and y coordinates for the center of the element.
        """
        self.get_element_coords(label, x, y, width, height, matching, best)
        return self.state.center()

    #########
    # Browser
    #########
    def page_title(self):
        """
        Returns the active page title.

        Returns:
            title (str): The page title.
        """
        try:
            return self._driver.title
        except InvalidSessionIdException:
            return None

    def page_source(self):
        """
        Returns the active page source.

        Returns:
            soup (BeautifulSoup): BeautifulSoup object for the page source.
        """
        try:
            soup = BeautifulSoup(self._driver.page_source, 'html.parser')
            return soup
        except InvalidSessionIdException:
            return None

    def navigate_to(self, url, is_retry=False):
        """
        Opens the browser on the given URL.

        Args:
            url (str):  The URL to be visited.
            is_retry (bool): Whether or not this is a retry attempt.
        """
        self._x = 0
        self._y = 0
        if not self._driver:
            self.start_browser()
        try:
            self._driver.get(url)
        except InvalidSessionIdException:
            if not is_retry:
                self.stop_browser()
                self.navigate_to(url, is_retry=True)

    def browse(self, url):
        """
        Opens the browser on the given URL.

        Args:
            url (str):  The URL to be visited.
        """
        self.navigate_to(url)

    def back(self):
        """
        Goes one step backward in the browser history.
        """
        self._driver.back()

    def forward(self):
        """
        Goes one step forward in the browser history.
        """
        self._driver.forward()

    def refresh(self):
        """
        Refreshes the current page.
        """
        self._driver.refresh()

    @contextmanager
    def wait_for_new_page(self, waiting_time=10000, activate=True):
        """Context manager to wait for a new page to load and activate it.

        Args:
            waiting_time (int, optional): The maximum waiting time. Defaults to 10000.
            activate (bool, optional): Whether or not to activate the new page. Defaults to True.

        """
        tabs = self.get_tabs()
        yield
        start_time = time.time()
        while tabs == self.get_tabs():
            elapsed_time = (time.time() - start_time) * 1000
            if elapsed_time > waiting_time:
                return None
            time.sleep(0.1)
        if activate:
            self.activate_tab(self.get_tabs()[-1])

    def execute_javascript(self, code):
        """
        Execute the given javascript code.

        Args:
            code (str): The code to be executed.

        Returns:
            value (object): Returns the code output or None if not available or if an error happens.
        """
        return self._driver.execute_script(code)

    def handle_js_dialog(self, accept=True, prompt_text=None):
        """
        Accepts or dismisses a JavaScript initiated dialog (alert, confirm, prompt, or onbeforeunload).
        This also cleans the dialog information in the local buffer.

        Args:
            accept (bool): Whether to accept or dismiss the dialog.
            prompt_text (str): The text to enter into the dialog prompt before accepting.
                Used only if this is a prompt dialog.
        """
        dialog = self.get_js_dialog()
        if not dialog:
            # TODO: Maybe we should raise an exception here if no alert available
            return
        if prompt_text is not None:
            dialog.send_keys(prompt_text)
        if accept:
            dialog.accept()
        else:
            dialog.dismiss()

    def get_js_dialog(self):
        """
        Return the last found dialog. Invoke first the `find_js_dialog` method to look up.

        Returns:
            dialog (dict): The dialog information or None if not available.
                See https://chromedevtools.github.io/devtools-protocol/tot/Page/#event-javascriptDialogOpening
        """
        try:
            dialog = self._driver.switch_to.alert
            return dialog
        except Exception:
            return None

    def get_tabs(self):
        """Get a list of tab handlers

        Returns:
            list: List of tab handlers
        """
        try:
            return self._driver.window_handles
        except InvalidSessionIdException:
            return []

    def create_tab(self, url):
        """Create a new tab and navigate to the given URL.

        Args:
            url (str): The desired URL.
        """
        try:
            # Refactor this when Selenium 4 is released
            self.execute_javascript(f"window.open('{url}', '_blank');")
            self._driver.switch_to.window(self.get_tabs()[-1])
        except InvalidSessionIdException:
            self.navigate_to(url)

    def create_window(self, url):
        """Creates a new window with the given URL.

        Args:
            url (str): The desired URL.
        """
        try:
            # Refactor this when Selenium 4 is released
            self.execute_javascript(f"window.open('{url}', '_blank', 'location=0');")
            self._driver.switch_to.window(self.get_tabs()[-1])
        except InvalidSessionIdException:
            self.navigate_to(url)

    def close_page(self):
        """Close the current active page (tab or window).
        """
        try:
            self._driver.close()

            # If it was the last tab we can't switch
            tabs = self.get_tabs()
            if tabs:
                self._driver.switch_to.window(tabs[-1])
        except InvalidSessionIdException:
            pass

    def activate_tab(self, handle):
        """Activate a tab given by the handle.

        Args:
            handle (str): The tab or window handle.
        """
        self._driver.switch_to.window(handle)

    def print_pdf(self, path=None, print_options=None):
        """Print the current page as a PDF file.

        Args:
            path (str, optional): The path for the file to be saved. Defaults to None.
            print_options (dict, optional): Print options as defined at. Defaults to None.

        Returns:
            str: the saved file path
        """
        title = self.page_title() or "document"
        title = re.sub("[\\\\|/:–]", "", title)
        timeout = 60000
        default_path = os.path.expanduser(os.path.join(self.download_folder_path, f"{title}.pdf"))

        if self.browser in [Browser.CHROME, Browser.EDGE] and not self.headless:
            pdf_current_count = self.get_file_count(file_extension=".pdf")
            # Chrome still does not support headless webdriver print
            # but Firefox does.
            self.execute_javascript("window.print();")

            # We need to wait for the file to be available in this case.
            if self.page_title():
                self.wait_for_file(default_path, timeout=timeout)
            else:
                # Waiting when the file don't have the page title in path
                self.wait_for_new_file(file_extension=".pdf", current_count=pdf_current_count)

            # Move the downloaded pdf file if the path is not None
            if path:
                last_downloaded_pdf = self.get_last_created_file(self.download_folder_path, ".pdf")
                os.rename(last_downloaded_pdf, path)
                return path
            self.wait(2000)
            return default_path

        if print_options is None:
            print_options = {
                'landscape': False,
                'displayHeaderFooter': False,
                'printBackground': True,
                'preferCSSPageSize': True,
                'marginTop': 0,
                'marginBottom': 0
            }
        data = self._webdriver_command("print", print_options)
        bytes_file = base64.b64decode(data)
        if not path:
            path = default_path
        with open(path, "wb") as f:
            f.write(bytes_file)
        return path

    def wait_for_downloads(self, timeout: int = 120000):
        """
        Wait for all downloads to be finished.
        Beware that this method replaces the current page with the downloads window.

        Args:
            timeout (int, optional): Timeout in millis. Defaults to 120000.
        """
        if self.browser in [Browser.CHROME, Browser.EDGE] and self.headless:
            start_time = time.time()
            while True:
                elapsed_time = (time.time() - start_time) * 1000
                if elapsed_time > timeout:
                    return False
                downloads_count = self.get_file_count(self.download_folder_path, ".crdownload")
                if downloads_count == 0:
                    return True
                self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)

        wait_method = BROWSER_CONFIGS.get(self.browser).get("wait_for_downloads")
        # waits for all the files to be completed
        WebDriverWait(self._driver, timeout/1000.0, 1).until(wait_method)

    def find_elements(self, selector: str, by: By = By.CSS_SELECTOR,
                      waiting_time=10000, ensure_visible: bool = True) -> List[WebElement]:
        """Find elements using the specified selector with selector type specified by `by`.

        Args:
            selector (str): The selector string to be used.
            by (str, optional): Selector type. Defaults to By.CSS_SELECTOR.
                [See more](https://selenium-python.readthedocs.io/api.html#selenium.webdriver.common.by.By)
            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
                Defaults to 10000ms (10s).
            ensure_visible (bool, optional): Whether to wait for the element to be visible. Defaults to True.

        Returns:
            List[WebElement]: List of elements found.

        **Example:**
        ```python
        from botcity.web import By
        ...
        # Find element by ID
        all_cells = self.find_elements("//td", By.XPATH)
        ...
        ```
        """
        if ensure_visible:
            condition = EC.visibility_of_all_elements_located
        else:
            condition = EC.presence_of_all_elements_located

        try:
            elements = WebDriverWait(
                self._driver, timeout=waiting_time / 1000.0
            ).until(
                condition((by, selector))
            )
            return elements
        except (TimeoutException, NoSuchElementException) as ex:
            print("Exception on find_elements", ex)
            return None

    def find_element(self, selector: str, by: str = By.CSS_SELECTOR, waiting_time=10000,
                     ensure_visible: bool = False, ensure_clickable: bool = False) -> WebElement:
        """Find an element using the specified selector with selector type specified by `by`.
        If more than one element is found, the first instance is returned.

        Args:
            selector (str): The selector string to be used.
            by (str, optional): Selector type. Defaults to By.CSS_SELECTOR.
                [See more](https://selenium-python.readthedocs.io/api.html#selenium.webdriver.common.by.By)
            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
                Defaults to 10000ms (10s).
            ensure_visible (bool, optional): Whether to wait for the element to be visible. Defaults to False.
            ensure_clickable (bool, optional): Whether to wait for the element to be clickable. Defaults to False.
                If True, `ensure_clickable` takes precedence over `ensure_visible`.

        Returns:
            WebElement: The element found.

        **Example:**
        ```python
        from botcity.web import By
        ...
        # Find element by ID
        elem = self.find_element("my_elem", By.ID)
        # Find element by XPath
        elem = self.find_element("//input[@type='submit']", By.XPATH)
        ...
        ```
        """
        condition = EC.visibility_of_element_located if ensure_visible else EC.presence_of_element_located
        condition = EC.element_to_be_clickable if ensure_clickable else condition

        try:
            element = WebDriverWait(
                self._driver, timeout=waiting_time/1000.0
            ).until(
                condition((by, selector))
            )
            return element
        except (TimeoutException, NoSuchElementException):
            return None

    def scroll_element(self, element: WebElement, steps: int = 100, interval: float = 500,
                       start: int = 0, end: int = None):
        """Scrolls down an element by its scroll height or a given amount defined by `start` and `end`.

        This is useful for scrolling down a page to load more content or
        to scroll down a dynamically loaded element.

        Args:
            element (WebElement): The element to scroll.
            steps (int, optional): Number of steps in which to conclude the scroll. Defaults to 100.
            interval (float, optional): Time interval between each step. Defaults to 500ms.
            start (int, optional): Start position. Defaults to 0.
            end (int, optional): End position. Defaults to None.
        """
        ele_height = self.driver.execute_script(
            "return arguments[0].scrollHeight;", element
        )

        start = max(0, start)
        end = min(ele_height, end) if end is not None else ele_height

        for i in range(start, end, steps):
            self.driver.execute_script(
                "arguments[0].scrollTo(0, arguments[1])", element, i)
            self.sleep(interval/1000.0)

    def wait_for_stale_element(self, element: WebElement, timeout: int = 10000):
        """
        Wait until the WebElement element becomes stale (outdated).

        Args:
            element (WebElement): The element to monitor for staleness.
            timeout (int, optional): Timeout in millis. Defaults to 120000.
        """
        try:
            WebDriverWait(self._driver, timeout=timeout/1000.0).until(EC.staleness_of(element))
        except (TimeoutException, NoSuchElementException):
            pass

    def wait_for_element_visibility(self, element: WebElement, visible: bool = True, waiting_time=10000):
        """Wait for the element to be visible or hidden.

        Args:
            element (WebElement): The element to wait for.
            visible (bool, optional): Whether to wait for the element to be visible. Defaults to True.
            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
                Defaults to 10000ms (10s).
        """
        if visible:
            wait_method = EC.visibility_of
        else:
            wait_method = EC.invisibility_of_element

        WebDriverWait(self._driver, timeout=waiting_time/1000.0).until(wait_method(element))

    def set_file_input_element(self, element: WebElement, filepath: str):
        """Configure the filepath for upload in a file element.
        Note: This method does not submit the form.

        Args:
            element (WebElement): The file upload element.
            filepath (str): The path to the file to be uploaded.

        **Example:**
        ```python
        ...
        # Find element
        elem = self.find_element("body > form > input[type=file]")
        # Set the filepath
        self.set_file_input_element(elem, "./test.txt")
        ...
        ```
        """
        fpath = os.path.abspath(os.path.expanduser(os.path.expandvars(filepath)))
        element.send_keys(fpath)

    def enter_iframe(self, iframe: WebElement):
        """Switch the WebBot driver to the specified iframe.

        Args:
            iframe (WebElement): The desired iFrame.
        """
        self._driver.switch_to.frame(iframe)

    def leave_iframe(self):
        """Leave the iframe and switch the WebBot driver to the default content.
        """
        self._driver.switch_to.default_content()

    def install_firefox_extension(self, extension):
        """
        Install an extension in the Firefox browser.
        This will start the browser if it was not started yet.

        Args:
            extension (str): The path of the .xpi extension to be loaded.
        """
        if self.browser != Browser.FIREFOX:
            raise ValueError("install_firefox_extension only works with Firefox.")
        if not self._driver:
            self.start_browser()
        self._driver.install_addon(os.path.abspath(extension))

    #######
    # Mouse
    #######
    def click_on(self, label):
        """
        Click on the element.

        Args:
            label (str): The image identifier
        """
        x, y = self.get_element_coords_centered(label)
        if None in (x, y):
            raise ValueError(f'Element not available. Cannot find {label}.')
        self.click_at(x, y)

    @only_if_element
    def get_last_x(self):
        """
        Get the last X position for the mouse.

        Returns:
            x (int): The last x position for the mouse.
        """
        return self._x

    def get_last_y(self):
        """
        Get the last Y position for the mouse.

        Returns:
            y (int): The last y position for the mouse.
        """
        return self._y

    def mouse_move(self, x, y):
        """
        Mouse the move to the coordinate defined by x and y

        Args:
            x (int): The X coordinate
            y (int): The Y coordinate

        """
        if self.browser == Browser.FIREFOX:
            # Reset coordinates if the page has gone stale. Only required for Firefox
            if self._html_elem is None:
                self._html_elem = self._driver.find_element_by_tag_name('body')
                self._x = 0
                self._y = 0
            else:
                try:
                    self._html_elem.is_enabled()
                except StaleElementReferenceException:
                    self._html_elem = self._driver.find_element_by_tag_name('body')
                    self._x = 0
                    self._y = 0

        mx = x - self._x
        my = y - self._y
        self._x = x
        self._y = y
        ActionChains(self._driver).move_by_offset(mx, my).perform()

    def click_at(self, x, y, *, clicks=1, interval_between_clicks=0, button='left'):
        """
        Click at the coordinate defined by x and y

        Args:
            x (int): The X coordinate
            y (int): The Y coordinate
            clicks (int, optional): Number of times to click. Defaults to 1.
            interval_between_clicks (int, optional): The interval between clicks in ms. Defaults to 0.
            button (str, optional): One of 'left', 'right'. Defaults to 'left'
        """
        self.mouse_move(x, y)
        ac = ActionChains(self._driver)
        for i in range(clicks):
            if button == 'left':
                ac.click()
            elif button == 'right':
                ac.context_click()
            else:
                raise ValueError('Invalid value for button. Accepted values are left or right.')
            ac.pause(interval_between_clicks/1000.0)
        ac.perform()

    @only_if_element
    def click(self, wait_after=config.DEFAULT_SLEEP_AFTER_ACTION, *,
              clicks=1, interval_between_clicks=0, button='left'):
        """
        Click on the last found element.

        Args:
            wait_after (int, optional): Interval to wait after clicking on the element.
            clicks (int, optional): Number of times to click. Defaults to 1.
            interval_between_clicks (int, optional): The interval between clicks in ms. Defaults to 0.
            button (str, optional): One of 'left', 'right'. Defaults to 'left'
        """
        x, y = self.state.center()
        self.click_at(x, y, clicks=clicks, button=button, interval_between_clicks=interval_between_clicks)
        self.sleep(wait_after)

    @only_if_element
    def click_relative(self, x, y, wait_after=config.DEFAULT_SLEEP_AFTER_ACTION, *,
                       clicks=1, interval_between_clicks=0, button='left'):
        """
        Click Relative on the last found element.

        Args:
            x (int): Horizontal offset
            y (int): Vertical offset
            wait_after (int, optional): Interval to wait after clicking on the element.
            clicks (int, optional): Number of times to click. Defaults to 1.
            interval_between_clicks (int, optional): The interval between clicks in ms. Defaults to 0.
            button (str, optional): One of 'left', 'right'. Defaults to 'left'
        """
        x = self.state.x() + x
        y = self.state.y() + y
        self.click_at(x, y, clicks=clicks, button=button, interval_between_clicks=interval_between_clicks)
        self.sleep(wait_after)

    @only_if_element
    def double_click(self, wait_after=config.DEFAULT_SLEEP_AFTER_ACTION):
        """
        Double Click on the last found element.

        Args:
            wait_after (int, optional): Interval to wait after clicking on the element.
        """
        self.click(interval_between_clicks=wait_after, clicks=2)

    @only_if_element
    def double_click_relative(self, x, y, interval_between_clicks=0, wait_after=config.DEFAULT_SLEEP_AFTER_ACTION):
        """
        Double Click Relative on the last found element.

        Args:
            x (int): Horizontal offset
            y (int): Vertical offset
            interval_between_clicks (int, optional): The interval between clicks in ms. Defaults to 0.
            wait_after (int, optional): Interval to wait after clicking on the element.
        """
        self.click_relative(x, y, wait_after=wait_after, clicks=2, interval_between_clicks=interval_between_clicks)

    @only_if_element
    def triple_click(self, wait_after=config.DEFAULT_SLEEP_AFTER_ACTION):
        """
        Triple Click on the last found element.

        Args:
            wait_after (int, optional): Interval to wait after clicking on the element.
        """
        self.click(wait_after=wait_after, clicks=3)

    @only_if_element
    def triple_click_relative(self, x, y, interval_between_clicks=0, wait_after=config.DEFAULT_SLEEP_AFTER_ACTION):
        """
        Triple Click Relative on the last found element.

        Args:
            x (int): Horizontal offset
            y (int): Vertical offset
            interval_between_clicks (int, optional): The interval between clicks in ms. Defaults to 0.
            wait_after (int, optional): Interval to wait after clicking on the element.
        """
        self.click_relative(x, y, wait_after=wait_after, clicks=3, interval_between_clicks=interval_between_clicks)

    def mouse_down(self, wait_after=config.DEFAULT_SLEEP_AFTER_ACTION, *, button='left'):
        """
        Holds down the requested mouse button.

        Args:
            wait_after (int, optional): Interval to wait after clicking on the element.
            button (str, optional): One of 'left', 'right', 'middle'. Defaults to 'left'
        """
        ActionChains(self._driver).click_and_hold().perform()
        self.sleep(wait_after)

    def mouse_up(self, wait_after=config.DEFAULT_SLEEP_AFTER_ACTION, *, button='left'):
        """
        Releases the requested mouse button.

        Args:
            wait_after (int, optional): Interval to wait after clicking on the element.
            button (str, optional): One of 'left', 'right', 'middle'. Defaults to 'left'
        """
        ActionChains(self._driver).release().perform()
        self.sleep(wait_after)

    def scroll_down(self, clicks):
        """
        Scroll Down n clicks

        Args:
            clicks (int): Number of times to scroll down.
        """
        for i in range(clicks):
            self._driver.execute_script("window.scrollTo(0, window.scrollY + 200)")
            self.sleep(200)

    def scroll_up(self, clicks):
        """
        Scroll Up n clicks

        Args:
            clicks (int): Number of times to scroll up.
        """
        for i in range(clicks):
            self._driver.execute_script("window.scrollTo(0, window.scrollY - 200)")
            self.sleep(200)

    def move_to(self, x, y):
        """
        Move the mouse relative to its current position.

        Args:
            x (int): The X coordinate
            y (int): The Y coordinate
        """
        self.mouse_move(x, y)

    @only_if_element
    def move(self):
        """
        Move to the center position of last found item.
        """
        x, y = self.state.center()
        self.move_to(x, y)

    def move_relative(self, x, y):
        """
        Move the mouse relative to its current position.

        Args:
            x (int): Horizontal offset
            y (int): Vertical offset

        """
        x = self.get_last_x() + x
        y = self.get_last_y() + y
        self.move_to(x, y)

    def move_random(self, range_x, range_y):
        """
        Move randomly along the given x, y range.

        Args:
            range_x (int): Horizontal range
            range_y (int): Vertical range

        """
        x = int(random.random() * range_x)
        y = int(random.random() * range_y)
        self.move_to(x, y)

    @only_if_element
    def right_click(self, wait_after=config.DEFAULT_SLEEP_AFTER_ACTION, *,
                    clicks=1, interval_between_clicks=0):
        """
        Right click on the last found element.

        Args:
            wait_after (int, optional): Interval to wait after clicking on the element.
            clicks (int, optional): Number of times to click. Defaults to 1.
            interval_between_clicks (int, optional): The interval between clicks in ms. Defaults to 0.
        """
        self.click(clicks=clicks, button='right', interval_between_clicks=interval_between_clicks)
        self.sleep(wait_after)

    def right_click_at(self, x, y):
        """
        Right click at the coordinate defined by x and y

        Args:
            x (int): The X coordinate
            y (int): The Y coordinate
        """
        self.click_at(x, y, button='right')

    @only_if_element
    def right_click_relative(self, x, y, interval_between_clicks=0, wait_after=config.DEFAULT_SLEEP_AFTER_ACTION):
        """
        Right Click Relative on the last found element.

        Args:
            x (int): Horizontal offset
            y (int): Vertical offset
            interval_between_clicks (int, optional): The interval between clicks in ms. Defaults to 0.
            wait_after (int, optional): Interval to wait after clicking on the element.

        """
        self.click_relative(x, y, wait_after=wait_after, interval_between_clicks=interval_between_clicks,
                            button='right')

    ##########
    # Keyboard
    ##########
    def kb_type(self, text, interval=0):
        """
        Type a text char by char (individual key events).

        Args:
            text (str): text to be typed.
            interval (int, optional): interval (ms) between each key press. Defaults to 0
        """
        action = ActionChains(self._driver)

        for c in text:
            action.send_keys(c)
            action.pause(interval / 1000.0)

        action.perform()
        self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)

    def paste(self, text=None, wait=0):
        """
        Paste content from the clipboard.

        Args:
            text (str, optional): The text to be pasted. Defaults to None
            wait (int, optional): Wait interval (ms) after task
        """
        text_to_paste = self._clipboard
        if text:
            text_to_paste = text
        self.kb_type(text_to_paste)

    def copy_to_clipboard(self, text, wait=0):
        """
        Copy content to the clipboard.

        Args:
            text (str): The text to be copied.
            wait (int, optional): Wait interval (ms) after task
        """
        self._clipboard = text
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def tab(self, wait=0):
        """
        Press key Tab

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.TAB)
        action.key_up(Keys.TAB)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def enter(self, wait=0):
        """
        Press key Enter

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.ENTER)
        action.key_up(Keys.ENTER)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def key_right(self, wait=0):
        """
        Press key Right

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.ARROW_RIGHT)
        action.key_up(Keys.ARROW_RIGHT)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def key_enter(self, wait=0):
        """
        Press key Enter

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        self.enter(wait)

    def key_home(self, wait=0):
        """
        Press key Home

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        # TODO: Investigate why with Firefox the key isn't working properly
        action = ActionChains(self._driver)
        action.key_down(Keys.HOME)
        action.key_up(Keys.HOME)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def key_end(self, wait=0):
        """
        Press key End

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.END)
        action.key_up(Keys.END)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def page_up(self, wait=0):
        """
        Press Page Up key

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        # TODO: Investigate why with Firefox the key isn't working properly
        action = ActionChains(self._driver)
        action.key_down(Keys.PAGE_UP)
        action.key_up(Keys.PAGE_UP)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def page_down(self, wait=0):
        """
        Press Page Down key

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        # TODO: Investigate why with Firefox the key isn't working properly
        action = ActionChains(self._driver)
        action.key_down(Keys.PAGE_DOWN)
        action.key_up(Keys.PAGE_DOWN)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def key_esc(self, wait=0):
        """
        Press key Esc

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.ESCAPE)
        action.key_up(Keys.ESCAPE)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def _key_fx(self, idx, wait=0):
        """
        Press key Fidx where idx is a value from 1 to 12

        Args:
            idx (int): F key index from 1 to 12
            wait (int, optional): Wait interval (ms) after task

        """
        if idx < 1 or idx > 12:
            raise ValueError("Only F1 to F12 allowed.")
        action = ActionChains(self._driver)
        key = getattr(Keys, f"F{idx}")
        action.key_down(key)
        action.key_up(key)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def hold_shift(self, wait=0):
        """
        Hold key Shift

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.SHIFT)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def release_shift(self):
        """
        Release key Shift.
        This method needs to be invoked after holding Shift or similar.
        """
        action = ActionChains(self._driver)
        action.key_up(Keys.SHIFT)
        action.perform()

    def maximize_window(self):
        """
        Shortcut to maximize window on Windows OS.
        """
        # TODO: Understand the complications associated with maximizing the browser and the resolution
        self._driver.maximize_window()

    def type_keys_with_interval(self, interval, keys):
        """
        Press a sequence of keys. Hold the keys in the specific order and releases them.

        Args:
            interval (int): Interval (ms) in which to press and release keys
            keys (list): List of Keys to be pressed
        """
        action = ActionChains(self._driver)

        for k in keys:
            action.key_down(k)
            action.pause(interval / 1000.0)
        for k in reversed(keys):
            action.key_up(k)
            action.pause(interval / 1000.0)
        action.perform()

    def type_keys(self, keys):
        """
        Press a sequence of keys. Hold the keys in the specific order and releases them.

        Args:
            keys (list): List of keys to be pressed
        """
        self.type_keys_with_interval(100, keys)

    def control_c(self, wait=0):
        """
        Press keys CTRL+C

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        # Firefox can't do window.getSelection() and return a proper value when the selected text
        # is in an input of similar. While Firefox doesn't get its shit together we apply this
        # ugly alternative so control+c works for "all" browsers tested so far.
        cmd = """
            try {
                return document.activeElement.value.substring(
                    document.activeElement.selectionStart,
                    document.activeElement.selectionEnd
                );
            } catch(error) {
                return window.getSelection().toString();
            }
        """

        self._clipboard = self.execute_javascript(cmd)
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def control_v(self, wait=0):
        """
        Press keys CTRL+V

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        self.paste()

    def control_a(self, wait=0):
        """
        Press keys CTRL+A

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        key = Keys.CONTROL
        if platform.system() == 'Darwin':
            key = Keys.COMMAND

        action.key_down(key)
        action.send_keys('a')
        action.key_up(key)
        action.perform()

    def get_clipboard(self):
        """
        Get the current content in the clipboard.

        Returns:
            text (str): Current clipboard content
        """
        return self._clipboard

    def type_left(self, wait=0):
        """
        Press Left key

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.ARROW_LEFT)
        action.key_up(Keys.ARROW_LEFT)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def type_right(self, wait=0):
        """
        Press Right key

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        self.key_right(wait=wait)

    def type_down(self, wait=0):
        """
        Press Down key

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.ARROW_DOWN)
        action.key_up(Keys.ARROW_DOWN)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def type_up(self, wait=0):
        """
        Press Up key

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.ARROW_UP)
        action.key_up(Keys.ARROW_UP)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def space(self, wait=0):
        """
        Press Space key

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.SPACE)
        action.key_up(Keys.SPACE)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def backspace(self, wait=0):
        """
        Press Backspace key

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.BACK_SPACE)
        action.key_up(Keys.BACK_SPACE)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    def delete(self, wait=0):
        """
        Press Delete key

        Args:
            wait (int, optional): Wait interval (ms) after task

        """
        action = ActionChains(self._driver)
        action.key_down(Keys.DELETE)
        action.key_up(Keys.DELETE)
        action.perform()
        delay = max(0, wait or config.DEFAULT_SLEEP_AFTER_ACTION)
        self.sleep(delay)

    ######
    # Misc
    ######

    def wait(self, interval):
        """
        Wait / Sleep for a given interval.

        Args:
            interval (int): Interval in milliseconds

        """
        time.sleep(interval / 1000.0)

    def sleep(self, interval):
        """
        Wait / Sleep for a given interval.

        Args:
            interval (int): Interval in milliseconds

        """
        self.wait(interval)

    def wait_for_file(self, path, timeout=60000):
        """
        Wait for a file to be available on disk.

        Args:
            path (str): The path for the file to be executed
            timeout (int, optional): Maximum wait time (ms) to search for a hit.
                Defaults to 60000ms (60s).

        Returns:
            status (bool): Whether or not the file was available before the timeout

        """
        path = os.path.abspath(os.path.expanduser(os.path.expandvars(path)))
        start_time = time.time()

        while True:
            elapsed_time = (time.time() - start_time) * 1000
            if elapsed_time > timeout:
                return False
            if os.path.isfile(path) and os.access(path, os.R_OK):
                if self.browser == Browser.FIREFOX and os.path.getsize(path) == 0:
                    # if file is empty, the download is not completed.
                    continue
                self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)
                return True
            self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)

    def get_last_created_file(self, path=None, file_extension=""):
        """Returns the last created file in a specific folder path.

        Args:
            path (str, optional): The path of the folder where the file is expected. Defaults to None.
            file_extension (str, optional): The extension of the file to be searched for (e.g., .pdf, .txt).

        Returns:
            str: the path of the last created file
        """
        if not path:
            path = self.download_folder_path

        files_path = glob.glob(os.path.expanduser(os.path.join(path, f"*{file_extension}")))
        last_created_file = max(files_path, key=os.path.getctime)
        return last_created_file

    def get_file_count(self, path=None, file_extension=""):
        """Get the total number of files of the same type.

        Args:
            path (str, optional): The path of the folder where the files are saved.
            file_extension (str, optional): The extension of the files to be searched for (e.g., .pdf, .txt).

        Returns:
            int: the number of files of the given type
        """
        if not path:
            path = self.download_folder_path

        files_path = glob.glob(os.path.expanduser(os.path.join(path, f"*{file_extension}")))
        return len(files_path)

    def wait_for_new_file(self, path=None, file_extension="", current_count=0, timeout=60000):
        """
        Wait for a new file to be available on disk without the file path.

        Args:
            path (str, optional): The path of the folder where the file is expected. Defaults to None.
            file_extension (str, optional): The extension of the file to be searched for (e.g., .pdf, .txt).
            current_count (int): The current number of files in the folder of the given type. Defaults to 0 files
            timeout (int, optional): Maximum wait time (ms) to search for a hit.
                Defaults to 60000ms (60s).

        Returns:
            str: the path of the last created file of the given type
        """
        if not path:
            path = self.download_folder_path

        start_time = time.time()
        while True:
            elapsed_time = (time.time() - start_time) * 1000
            if elapsed_time > timeout:
                return None
            file_count = self.get_file_count(path, f"*{file_extension}")
            if file_count == current_count + 1:
                self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)
                return self.get_last_created_file(path, f"*{file_extension}")
            self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)



import atexit
import os
import tempfile
from typing import Dict

from selenium import webdriver
from selenium.webdriver import Firefox  # noqa: F401, F403
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.firefox.options import Options as FirefoxOptions

from ..util import cleanup_temp_dir

FIREFOX_MIMETYPES_TO_DOWNLOAD = ['application/vnd.hzn-3d-crossword', 'video/3gpp', 'video/3gpp2',
                                 'application/vnd.mseq', 'application/vnd.3m.post-it-notes',
                                 'application/vnd.3gpp.pic-bw-large', 'application/vnd.3gpp.pic-bw-small',
                                 'application/vnd.3gpp.pic-bw-var', 'application/vnd.3gp2.tcap',
                                 'application/x-7z-compressed', 'application/x-abiword',
                                 'application/x-ace-compressed',
                                 'application/vnd.americandynamics.acc', 'application/vnd.acucobol',
                                 'application/vnd.acucorp', 'audio/adpcm', 'application/x-authorware-bin',
                                 'application/x-athorware-map', 'application/x-authorware-seg',
                                 'application/vnd.adobe.air-application-installer-package+zip',
                                 'application/x-shockwave-flash', 'application/vnd.adobe.fxp', 'application/pdf',
                                 'application/vnd.cups-ppd', 'application/x-director',
                                 'applicaion/vnd.adobe.xdp+xml',
                                 'application/vnd.adobe.xfdf', 'audio/x-aac', 'application/vnd.ahead.space',
                                 'application/vnd.airzip.filesecure.azf', 'application/vnd.airzip.filesecure.azs',
                                 'application/vnd.amazon.ebook', 'application/vnd.amiga.ami',
                                 'applicatin/andrew-inset',
                                 'application/vnd.android.package-archive',
                                 'application/vnd.anser-web-certificate-issue-initiation',
                                 'application/vnd.anser-web-funds-transfer-initiation',
                                 'application/vnd.antix.game-component', 'application/vnd.apple.installe+xml',
                                 'application/applixware', 'application/vnd.hhe.lesson-player',
                                 'application/vnd.aristanetworks.swi', 'text/x-asm', 'application/atomcat+xml',
                                 'application/atomsvc+xml', 'application/atom+xml', 'application/pkix-attr-cert',
                                 'audio/x-aiff', 'video/x-msvieo', 'application/vnd.audiograph', 'image/vnd.dxf',
                                 'model/vnd.dwf', 'text/plain-bas', 'application/x-bcpio',
                                 'application/octet-stream',
                                 'image/bmp', 'application/x-bittorrent', 'application/vnd.rim.cod',
                                 'application/vnd.blueice.multipass', 'application/vnd.bm', 'application/x-sh',
                                 'image/prs.btif', 'application/vnd.businessobjects', 'application/x-bzip',
                                 'application/x-bzip2', 'application/x-csh', 'text/x-c',
                                 'application/vnd.chemdraw+xml',
                                 'text/css', 'chemical/x-cdx', 'chemical/x-cml', 'chemical/x-csml',
                                 'application/vn.contact.cmsg', 'application/vnd.claymore',
                                 'application/vnd.clonk.c4group', 'image/vnd.dvb.subtitle',
                                 'application/cdmi-capability', 'application/cdmi-container',
                                 'application/cdmi-domain',
                                 'application/cdmi-object', 'application/cdmi-queue',
                                 'applicationvnd.cluetrust.cartomobile-config',
                                 'application/vnd.cluetrust.cartomobile-config-pkg', 'image/x-cmu-raster',
                                 'model/vnd.collada+xml', 'text/csv', 'application/mac-compactpro',
                                 'application/vnd.wap.wmlc', 'image/cgm', 'x-conference/x-cooltalk', 'image/x-cmx',
                                 'application/vnd.xara', 'application/vnd.cosmocaller', 'application/x-cpio',
                                 'application/vnd.crick.clicker', 'application/vnd.crick.clicker.keyboard',
                                 'application/vnd.crick.clicker.palette', 'application/vnd.crick.clicker.template',
                                 'application/vn.crick.clicker.wordbank', 'application/vnd.criticaltools.wbs+xml',
                                 'application/vnd.rig.cryptonote', 'chemical/x-cif', 'chemical/x-cmdf',
                                 'application/cu-seeme', 'application/prs.cww', 'text/vnd.curl',
                                 'text/vnd.curl.dcurl',
                                 'text/vnd.curl.mcurl', 'text/vnd.crl.scurl', 'application/vnd.curl.car',
                                 'application/vnd.curl.pcurl', 'application/vnd.yellowriver-custom-menu',
                                 'application/dssc+der', 'application/dssc+xml', 'application/x-debian-package',
                                 'audio/vnd.dece.audio', 'image/vnd.dece.graphic', 'video/vnd.dec.hd',
                                 'video/vnd.dece.mobile', 'video/vnd.uvvu.mp4', 'video/vnd.dece.pd',
                                 'video/vnd.dece.sd', 'video/vnd.dece.video', 'application/x-dvi',
                                 'application/vnd.fdsn.seed', 'application/x-dtbook+xml',
                                 'application/x-dtbresource+xml', 'application/vnd.dvb.ait',
                                 'applcation/vnd.dvb.service', 'audio/vnd.digital-winds', 'image/vnd.djvu',
                                 'application/xml-dtd', 'application/vnd.dolby.mlp', 'application/x-doom',
                                 'application/vnd.dpgraph', 'audio/vnd.dra', 'application/vnd.dreamfactory',
                                 'audio/vnd.dts', 'audio/vnd.dts.hd', 'imag/vnd.dwg', 'application/vnd.dynageo',
                                 'application/ecmascript', 'application/vnd.ecowin.chart',
                                 'image/vnd.fujixerox.edmics-mmr', 'image/vnd.fujixerox.edmics-rlc',
                                 'application/exi',
                                 'application/vnd.proteus.magazine', 'application/epub+zip', 'message/rfc82',
                                 'application/vnd.enliven', 'application/vnd.is-xpr', 'image/vnd.xiff',
                                 'application/vnd.xfdl', 'application/emma+xml', 'application/vnd.ezpix-album',
                                 'application/vnd.ezpix-package', 'image/vnd.fst', 'video/vnd.fvt',
                                 'image/vnd.fastbidsheet', 'application/vn.denovo.fcselayout-link', 'video/x-f4v',
                                 'video/x-flv', 'image/vnd.fpx', 'image/vnd.net-fpx', 'text/vnd.fmi.flexstor',
                                 'video/x-fli', 'application/vnd.fluxtime.clip', 'application/vnd.fdf',
                                 'text/x-fortran', 'application/vnd.mif', 'application/vnd.framemaker',
                                 'imae/x-freehand', 'application/vnd.fsc.weblaunch', 'application/vnd.frogans.fnc',
                                 'application/vnd.frogans.ltf', 'application/vnd.fujixerox.ddd',
                                 'application/vnd.fujixerox.docuworks',
                                 'application/vnd.fujixerox.docuworks.binder',
                                 'application/vnd.fujitu.oasys', 'application/vnd.fujitsu.oasys2',
                                 'application/vnd.fujitsu.oasys3', 'application/vnd.fujitsu.oasysgp',
                                 'application/vnd.fujitsu.oasysprs', 'application/x-futuresplash',
                                 'application/vnd.fuzzysheet', 'image/g3fax', 'application/vnd.gmx',
                                 'model/vn.gtw',
                                 'application/vnd.genomatix.tuxedo', 'application/vnd.geogebra.file',
                                 'application/vnd.geogebra.tool', 'model/vnd.gdl',
                                 'application/vnd.geometry-explorer',
                                 'application/vnd.geonext', 'application/vnd.geoplan', 'application/vnd.geospace',
                                 'applicatio/x-font-ghostscript', 'application/x-font-bdf', 'application/x-gtar',
                                 'application/x-texinfo', 'application/x-gnumeric',
                                 'application/vnd.google-earth.kml+xml', 'application/vnd.google-earth.kmz',
                                 'application/vnd.grafeq', 'image/gif', 'text/vnd.graphviz',
                                 'aplication/vnd.groove-account', 'application/vnd.groove-help',
                                 'application/vnd.groove-identity-message', 'application/vnd.groove-injector',
                                 'application/vnd.groove-tool-message', 'application/vnd.groove-tool-template',
                                 'application/vnd.groove-vcar', 'video/h261', 'video/h263', 'video/h264',
                                 'application/vnd.hp-hpid', 'application/vnd.hp-hps', 'application/x-hdf',
                                 'audio/vnd.rip', 'application/vnd.hbci', 'application/vnd.hp-jlyt',
                                 'application/vnd.hp-pcl', 'application/vnd.hp-hpgl',
                                 'application/vnd.yamaha.h-script',
                                 'application/vnd.yamaha.hv-dic', 'application/vnd.yamaha.hv-voice',
                                 'application/vnd.hydrostatix.sof-data', 'application/hyperstudio',
                                 'application/vnd.hal+xml', 'text/html', 'application/vnd.ibm.rights-management',
                                 'application/vnd.ibm.securecontainer', 'text/calendar',
                                 'application/vnd.iccprofile',
                                 'image/x-icon', 'application/vnd.igloader', 'image/ief',
                                 'application/vnd.immervision-ivp', 'application/vnd.immervision-ivu',
                                 'application/reginfo+xml', 'text/vnd.in3d.3dml', 'text/vnd.in3d.spot',
                                 'mode/iges',
                                 'application/vnd.intergeo', 'application/vnd.cinderella',
                                 'application/vnd.intercon.formnet', 'application/vnd.isac.fcs',
                                 'application/ipfix',
                                 'application/pkix-cert', 'application/pkixcmp', 'application/pkix-crl',
                                 'application/pkix-pkipath', 'applicaion/vnd.insors.igm',
                                 'application/vnd.ipunplugged.rcprofile',
                                 'application/vnd.irepository.package+xml',
                                 'text/vnd.sun.j2me.app-descriptor', 'application/java-archive',
                                 'application/java-vm',
                                 'application/x-java-jnlp-file', 'application/java-serializd-object',
                                 'text/x-java-source', 'java', 'application/javascript', 'application/json',
                                 'application/vnd.joost.joda-archive', 'video/jpm', 'image/jpeg', 'video/jpeg',
                                 'application/vnd.kahootz', 'application/vnd.chipnuts.karaoke-mmd',
                                 'application/vnd.kde.karbon', 'aplication/vnd.kde.kchart',
                                 'application/vnd.kde.kformula', 'application/vnd.kde.kivio',
                                 'application/vnd.kde.kontour', 'application/vnd.kde.kpresenter',
                                 'application/vnd.kde.kspread', 'application/vnd.kde.kword',
                                 'application/vnd.kenameaapp', 'applicatin/vnd.kidspiration',
                                 'application/vnd.kinar',
                                 'application/vnd.kodak-descriptor', 'application/vnd.las.las+xml',
                                 'application/x-latex', 'application/vnd.llamagraphics.life-balance.desktop',
                                 'application/vnd.llamagraphics.life-balance.exchange+xml', 'application/vnd.jam',
                                 'application/vnd.lotus-1-2-3', 'application/vnd.lotus-approach',
                                 'application/vnd.lotus-freelance', 'application/vnd.lotus-notes',
                                 'application/vnd.lotus-organizer', 'application/vnd.lotus-screencam',
                                 'application/vnd.lotus-wordro', 'audio/vnd.lucent.voice', 'audio/x-mpegurl',
                                 'video/x-m4v', 'application/mac-binhex40', 'application/vnd.macports.portpkg',
                                 'application/vnd.osgeo.mapguide.package', 'application/marc',
                                 'application/marcxml+xml', 'application/mxf', 'application/vnd.wolfrm.player',
                                 'application/mathematica', 'application/mathml+xml', 'application/mbox',
                                 'application/vnd.medcalcdata', 'application/mediaservercontrol+xml',
                                 'application/vnd.mediastation.cdkey', 'application/vnd.mfer',
                                 'application/vnd.mfmp',
                                 'model/mesh', 'appliation/mads+xml', 'application/mets+xml',
                                 'application/mods+xml',
                                 'application/metalink4+xml',
                                 'application/vnd.ms-powerpoint.template.macroenabled.12',
                                 'application/vnd.ms-word.document.macroenabled.12',
                                 'application/vnd.ms-word.template.macroenabed.12', 'application/vnd.mcd',
                                 'application/vnd.micrografx.flo', 'application/vnd.micrografx.igx',
                                 'application/vnd.eszigno3+xml', 'application/x-msaccess', 'video/x-ms-asf',
                                 'application/x-msdownload', 'application/vnd.ms-artgalry',
                                 'application/vnd.ms-ca-compressed', 'application/vnd.ms-ims',
                                 'application/x-ms-application', 'application/x-msclip', 'image/vnd.ms-modi',
                                 'application/vnd.ms-fontobject', 'application/vnd.ms-excel',
                                 'application/vnd.ms-excel.addin.macroenabled.12',
                                 'application/vnd.ms-excelsheet.binary.macroenabled.12',
                                 'application/vnd.ms-excel.template.macroenabled.12',
                                 'application/vnd.ms-excel.sheet.macroenabled.12', 'application/vnd.ms-htmlhelp',
                                 'application/x-mscardfile', 'application/vnd.ms-lrm', 'application/x-msmediaview',
                                 'aplication/x-msmoney',
                                 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
                                 'application/vnd.openxmlformats-officedocument.presentationml.slide',
                                 'application/vnd.openxmlformats-officedocument.presentationml.slideshw',
                                 'application/vnd.openxmlformats-officedocument.presentationml.template',
                                 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
                                 'application/vnd.openxmlformats-officedocument.spreadsheetml.template',
                                 'application/vnd.openxmformats-officedocument.wordprocessingml.document',
                                 'application/vnd.openxmlformats-officedocument.wordprocessingml.template',
                                 'application/x-msbinder', 'application/vnd.ms-officetheme', 'application/onenote',
                                 'audio/vnd.ms-playready.media.pya', 'vdeo/vnd.ms-playready.media.pyv',
                                 'application/vnd.ms-powerpoint',
                                 'application/vnd.ms-powerpoint.addin.macroenabled.12',
                                 'application/vnd.ms-powerpoint.slide.macroenabled.12',
                                 'application/vnd.ms-powerpoint.presentation.macroenabled.12',
                                 'appliation/vnd.ms-powerpoint.slideshow.macroenabled.12',
                                 'application/vnd.ms-project',
                                 'application/x-mspublisher', 'application/x-msschedule',
                                 'application/x-silverlight-app', 'application/vnd.ms-pki.stl',
                                 'application/vnd.ms-pki.seccat', 'application/vn.visio', 'video/x-ms-wm',
                                 'audio/x-ms-wma', 'audio/x-ms-wax', 'video/x-ms-wmx', 'application/x-ms-wmd',
                                 'application/vnd.ms-wpl', 'application/x-ms-wmz', 'video/x-ms-wmv',
                                 'video/x-ms-wvx',
                                 'application/x-msmetafile', 'application/x-msterminal', 'application/msword',
                                 'application/x-mswrite', 'application/vnd.ms-works', 'application/x-ms-xbap',
                                 'application/vnd.ms-xpsdocument', 'audio/midi', 'application/vnd.ibm.minipay',
                                 'application/vnd.ibm.modcap', 'application/vnd.jcp.javame.midlet-rms',
                                 'application/vnd.tmobile-ivetv', 'application/x-mobipocket-ebook',
                                 'application/vnd.mobius.mbk', 'application/vnd.mobius.dis',
                                 'application/vnd.mobius.plc', 'application/vnd.mobius.mqy',
                                 'application/vnd.mobius.msl', 'application/vnd.mobius.txf',
                                 'application/vnd.mobius.daf', 'tex/vnd.fly', 'application/vnd.mophun.certificate',
                                 'application/vnd.mophun.application', 'video/mj2', 'audio/mpeg',
                                 'video/vnd.mpegurl',
                                 'video/mpeg', 'application/mp21', 'audio/mp4', 'video/mp4', 'application/mp4',
                                 'application/vnd.apple.mpegurl', 'application/vnd.msician',
                                 'application/vnd.muvee.style', 'application/xv+xml',
                                 'application/vnd.nokia.n-gage.data',
                                 'application/vnd.nokia.n-gage.symbian.install',
                                 'application/x-dtbncx+xml', 'application/x-netcdf',
                                 'application/vnd.neurolanguage.nlu', 'application/vnd.na',
                                 'application/vnd.noblenet-directory', 'application/vnd.noblenet-sealer',
                                 'application/vnd.noblenet-web', 'application/vnd.nokia.radio-preset',
                                 'application/vnd.nokia.radio-presets', 'text/n3', 'application/vnd.novadigm.edm',
                                 'application/vnd.novadim.edx', 'application/vnd.novadigm.ext',
                                 'application/vnd.flographit', 'audio/vnd.nuera.ecelp4800',
                                 'audio/vnd.nuera.ecelp7470',
                                 'audio/vnd.nuera.ecelp9600', 'application/oda', 'application/ogg', 'audio/ogg',
                                 'video/ogg', 'application/vnd.oma.dd2+xml',
                                 'applicatin/vnd.oasis.opendocument.text-web', 'application/oebps-package+xml',
                                 'application/vnd.intu.qbo', 'application/vnd.openofficeorg.extension',
                                 'application/vnd.yamaha.openscoreformat', 'audio/webm', 'video/webm',
                                 'application/vnd.oasis.opendocument.char',
                                 'application/vnd.oasis.opendocument.chart-template',
                                 'application/vnd.oasis.opendocument.database',
                                 'application/vnd.oasis.opendocument.formula',
                                 'application/vnd.oasis.opendocument.formula-template',
                                 'application/vnd.oasis.opendocument.grapics',
                                 'application/vnd.oasis.opendocument.graphics-template',
                                 'application/vnd.oasis.opendocument.image',
                                 'application/vnd.oasis.opendocument.image-template',
                                 'application/vnd.oasis.opendocument.presentation',
                                 'application/vnd.oasis.opendocumen.presentation-template',
                                 'application/vnd.oasis.opendocument.spreadsheet',
                                 'application/vnd.oasis.opendocument.spreadsheet-template',
                                 'application/vnd.oasis.opendocument.text',
                                 'application/vnd.oasis.opendocument.text-master',
                                 'application/vnd.asis.opendocument.text-template', 'image/ktx',
                                 'application/vnd.sun.xml.calc', 'application/vnd.sun.xml.calc.template',
                                 'application/vnd.sun.xml.draw', 'application/vnd.sun.xml.draw.template',
                                 'application/vnd.sun.xml.impress', 'application/vnd.sun.xl.impress.template',
                                 'application/vnd.sun.xml.math', 'application/vnd.sun.xml.writer',
                                 'application/vnd.sun.xml.writer.global',
                                 'application/vnd.sun.xml.writer.template',
                                 'application/x-font-otf', 'application/vnd.yamaha.openscoreformat.osfpvg+xml',
                                 'application/vnd.osgi.dp', 'application/vnd.palm', 'text/x-pascal',
                                 'application/vnd.pawaafile', 'application/vnd.hp-pclxl', 'application/vnd.picsel',
                                 'image/x-pcx', 'image/vnd.adobe.photoshop', 'application/pics-rules',
                                 'image/x-pict',
                                 'application/x-chat', 'aplication/pkcs10', 'application/x-pkcs12',
                                 'application/pkcs7-mime', 'application/pkcs7-signature',
                                 'application/x-pkcs7-certreqresp', 'application/x-pkcs7-certificates',
                                 'application/pkcs8', 'application/vnd.pocketlearn', 'image/x-portable-anymap',
                                 'image/-portable-bitmap', 'application/x-font-pcf', 'application/font-tdpfr',
                                 'application/x-chess-pgn', 'image/x-portable-graymap', 'image/png',
                                 'image/x-portable-pixmap', 'application/pskc+xml', 'application/vnd.ctc-posml',
                                 'application/postscript', 'application/xfont-type1',
                                 'application/vnd.powerbuilder6',
                                 'application/pgp-encrypted', 'application/pgp-signature',
                                 'application/vnd.previewsystems.box', 'application/vnd.pvi.ptid1',
                                 'application/pls+xml', 'application/vnd.pg.format', 'application/vnd.pg.osasli',
                                 'tex/prs.lines.tag', 'application/x-font-linux-psf',
                                 'application/vnd.publishare-delta-tree', 'application/vnd.pmi.widget',
                                 'application/vnd.quark.quarkxpress', 'application/vnd.epson.esf',
                                 'application/vnd.epson.msf', 'application/vnd.epson.ssf',
                                 'applicaton/vnd.epson.quickanime', 'application/vnd.intu.qfx', 'video/quicktime',
                                 'application/x-rar-compressed', 'audio/x-pn-realaudio',
                                 'audio/x-pn-realaudio-plugin',
                                 'application/rsd+xml', 'application/vnd.rn-realmedia',
                                 'application/vnd.realvnc.bed',
                                 'applicatin/vnd.recordare.musicxml', 'application/vnd.recordare.musicxml+xml',
                                 'application/relax-ng-compact-syntax', 'application/vnd.data-vision.rdz',
                                 'application/rdf+xml', 'application/vnd.cloanto.rp9', 'application/vnd.jisp',
                                 'application/rtf', 'text/richtex', 'application/vnd.route66.link66+xml',
                                 'application/rss+xml', 'application/shf+xml',
                                 'application/vnd.sailingtracker.track',
                                 'image/svg+xml', 'application/vnd.sus-calendar', 'application/sru+xml',
                                 'application/set-payment-initiation', 'application/set-reistration-initiation',
                                 'application/vnd.sema', 'application/vnd.semd', 'application/vnd.semf',
                                 'application/vnd.seemail', 'application/x-font-snf',
                                 'application/scvp-vp-request',
                                 'application/scvp-vp-response', 'application/scvp-cv-request',
                                 'application/svp-cv-response', 'application/sdp', 'text/x-setext',
                                 'video/x-sgi-movie',
                                 'application/vnd.shana.informed.formdata',
                                 'application/vnd.shana.informed.formtemplate',
                                 'application/vnd.shana.informed.interchange',
                                 'application/vnd.shana.informed.package',
                                 'application/thraud+xml', 'application/x-shar', 'image/x-rgb',
                                 'application/vnd.epson.salt', 'application/vnd.accpac.simply.aso',
                                 'application/vnd.accpac.simply.imp', 'application/vnd.simtech-mindmapper',
                                 'application/vnd.commonspace', 'application/vnd.ymaha.smaf-audio',
                                 'application/vnd.smaf', 'application/vnd.yamaha.smaf-phrase',
                                 'application/vnd.smart.teacher', 'application/vnd.svd',
                                 'application/sparql-query',
                                 'application/sparql-results+xml', 'application/srgs', 'application/srgs+xml',
                                 'application/sml+xml', 'application/vnd.koan', 'text/sgml',
                                 'application/vnd.stardivision.calc', 'application/vnd.stardivision.draw',
                                 'application/vnd.stardivision.impress', 'application/vnd.stardivision.math',
                                 'application/vnd.stardivision.writer',
                                 'application/vnd.tardivision.writer-global',
                                 'application/vnd.stepmania.stepchart', 'application/x-stuffit',
                                 'application/x-stuffitx', 'application/vnd.solent.sdkm+xml',
                                 'application/vnd.olpc-sugar', 'audio/basic', 'application/vnd.wqd',
                                 'application/vnd.symbian.install', 'application/smil+xml',
                                 'application/vnd.syncml+xml', 'application/vnd.syncml.dm+wbxml',
                                 'application/vnd.syncml.dm+xml', 'application/x-sv4cpio', 'application/x-sv4crc',
                                 'application/sbml+xml', 'text/tab-separated-values', 'image/tiff',
                                 'application/vnd.to.intent-module-archive', 'application/x-tar',
                                 'application/x-tcl',
                                 'application/x-tex', 'application/x-tex-tfm', 'application/tei+xml', 'text/plain',
                                 'application/vnd.spotfire.dxp', 'application/vnd.spotfire.sfs',
                                 'application/timestamped-data', 'applicationvnd.trid.tpt',
                                 'application/vnd.triscape.mxs', 'text/troff', 'application/vnd.trueapp',
                                 'application/x-font-ttf', 'text/turtle', 'application/vnd.umajin',
                                 'application/vnd.uoml+xml', 'application/vnd.unity', 'application/vnd.ufdl',
                                 'text/uri-list', 'application/nd.uiq.theme', 'application/x-ustar',
                                 'text/x-uuencode',
                                 'text/x-vcalendar', 'text/x-vcard', 'application/x-cdlink', 'application/vnd.vsf',
                                 'model/vrml', 'application/vnd.vcx', 'model/vnd.mts', 'model/vnd.vtu',
                                 'application/vnd.visionary', 'video/vnd.vivo', 'applicatin/ccxml+xml', '',
                                 'application/voicexml+xml', 'application/x-wais-source',
                                 'application/vnd.wap.wbxml',
                                 'image/vnd.wap.wbmp', 'audio/x-wav', 'application/davmount+xml',
                                 'application/x-font-woff', 'application/wspolicy+xml', 'image/webp',
                                 'application/vnd.webturb', 'application/widget', 'application/winhlp',
                                 'text/vnd.wap.wml', 'text/vnd.wap.wmlscript', 'application/vnd.wap.wmlscriptc',
                                 'application/vnd.wordperfect', 'application/vnd.wt.stf', 'application/wsdl+xml',
                                 'image/x-xbitmap', 'image/x-xpixmap', 'image/x-xwindowump',
                                 'application/x-x509-ca-cert', 'application/x-xfig', 'application/xhtml+xml',
                                 'application/xml', 'application/xcap-diff+xml', 'application/xenc+xml',
                                 'application/patch-ops-error+xml', 'application/resource-lists+xml',
                                 'application/rls-services+xml', 'aplication/resource-lists-diff+xml',
                                 'application/xslt+xml', 'application/xop+xml', 'application/x-xpinstall',
                                 'application/xspf+xml', 'application/vnd.mozilla.xul+xml', 'chemical/x-xyz',
                                 'text/yaml', 'application/yang', 'application/yin+xml', 'application/vnd.ul',
                                 'application/zip', 'application/vnd.handheld-entertainment+xml',
                                 'application/vnd.zzazz.deck+xml']


def default_options(headless=False, download_folder_path=None, user_data_dir=None,
                    page_load_strategy="normal") -> FirefoxOptions:
    """Retrieve the default options for this browser curated by BotCity.

    Args:
        headless (bool, optional): Whether or not to use the headless mode. Defaults to False.
        download_folder_path (str, optional): The default path in which to save files.
            If None, the current directory is used. Defaults to None.
        user_data_dir ([type], optional): The directory to use as user profile.
            If None, a new temporary directory is used. Defaults to None.
        page_load_strategy (str, optional): The page load strategy to use.

    Returns:
        FirefoxOptions: The Firefox options.
    """
    firefox_options = FirefoxOptions()
    try:
        page_load_strategy = page_load_strategy.value
    except AttributeError:
        page_load_strategy = page_load_strategy
    firefox_options.page_load_strategy = page_load_strategy
    firefox_options.headless = headless
    if not user_data_dir:
        temp_dir = tempfile.TemporaryDirectory(prefix="botcity_")
        user_data_dir = temp_dir.name
        atexit.register(cleanup_temp_dir, temp_dir)
    firefox_profile = webdriver.FirefoxProfile(user_data_dir)
    firefox_profile.set_preference("security.default_personal_cert", "Select Automatically")
    firefox_profile.set_preference('browser.download.folderList', 2)
    firefox_profile.set_preference('browser.download.manager.showWhenStarting', False)
    if not download_folder_path:
        download_folder_path = os.getcwd()
    firefox_profile.set_preference('browser.download.dir', download_folder_path)
    firefox_profile.set_preference('general.warnOnAboutConfig', False)

    mimetypes_to_download = ",".join(FIREFOX_MIMETYPES_TO_DOWNLOAD)
    firefox_profile.set_preference("pdfjs.disabled", True)
    firefox_profile.set_preference("plugin.disable_full_page_plugin_for_types", mimetypes_to_download)
    firefox_profile.set_preference('browser.helperApps.neverAsk.saveToDisk', mimetypes_to_download)

    firefox_profile.update_preferences()
    firefox_options.profile = firefox_profile

    return firefox_options


def default_capabilities() -> Dict:
    """Fetch the default capabilities for this browser.

    Returns:
        Dict: Dictionary with the default capabilities defined.
    """
    return DesiredCapabilities.FIREFOX.copy()


def wait_for_downloads(driver):
    """Wait for all downloads to finish.
    *Important*: This method overwrites the current page with the downloads page.
    """
    if not driver.current_url.startswith("about:downloads"):
        driver.get("about:downloads")

    return driver.execute_script("""
        var items = document.querySelector('richlistbox').itemChildren;
        if (items.every(e => e.attributes["state"].value != 0))
            return true;
        """)



import enum

from . import chrome
from . import firefox
from . import edge
from . import ie


class Browser(str, enum.Enum):
    """
    Supported Driver.

    Attributes:
        CHROME (str): Google Chrome
        FIREFOX (str): Mozilla Firefox
        EDGE (str): Microsoft Edge
        IE (str): Microsoft Internet Explorer
    """
    CHROME = "chrome"
    FIREFOX = "firefox"
    EDGE = "edge"
    IE = "ie"


class PageLoadStrategy(str, enum.Enum):
    """
    Page Load Strategy.

    Attributes:
        NORMAL (str): Wait for the entire page is loaded. When set to normal,
            waits until the load event fire is returned.
        EAGER (str): Wait until the initial HTML document has been completely
            loaded and parsed, and discards loading of stylesheets, images and subframes.
        NONE (str): Only waits until the initial page is downloaded
    """
    NORMAL = "normal"
    EAGER = "eager"
    NONE = "none"


BROWSER_CONFIGS = {
    Browser.CHROME: {
        "driver": "chromedriver",
        "class": chrome.Chrome,
        "options": chrome.default_options,
        "capabilities": chrome.default_capabilities,
        "wait_for_downloads": chrome.wait_for_downloads
    },
    Browser.FIREFOX: {
        "driver": "geckodriver",
        "class": firefox.Firefox,
        "options": firefox.default_options,
        "capabilities": firefox.default_capabilities,
        "wait_for_downloads": firefox.wait_for_downloads
    },
    Browser.EDGE: {
        "driver": "msedgedriver",
        "class": edge.Edge,
        "options": edge.default_options,
        "capabilities": edge.default_capabilities,
        "wait_for_downloads": edge.wait_for_downloads
    },
    Browser.IE: {
        "driver": "IEDriverServer",
        "class": ie.Ie,
        "options": ie.default_options,
        "capabilities": ie.default_capabilities,
        "wait_for_downloads": ie.wait_for_downloads
    },
}



"""
This module was extracted from https://github.com/asweigart/pyscreeze and modified to fit with our search when
using the web bot and also for PEP8 compliance.

Copyright (c) 2014, Al Sweigart
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of PyScreeze nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import collections
import cv2
import numpy

RUNNING_CV_2 = cv2.__version__[0] < '3'

Box = collections.namedtuple('Box', 'left top width height')

if RUNNING_CV_2:
    LOAD_COLOR = cv2.CV_LOAD_IMAGE_COLOR
    LOAD_GRAYSCALE = cv2.CV_LOAD_IMAGE_GRAYSCALE
else:
    LOAD_COLOR = cv2.IMREAD_COLOR
    LOAD_GRAYSCALE = cv2.IMREAD_GRAYSCALE


def _load_cv2(img, grayscale=False):
    """
    TODO
    """
    # load images if given filename, or convert as needed to opencv
    # Alpha layer just causes failures at this point, so flatten to RGB.
    # RGBA: load with -1 * cv2.CV_LOAD_IMAGE_COLOR to preserve alpha
    # to matchTemplate, need template and image to be the same wrt having alpha

    if isinstance(img, str):
        # The function imread loads an image from the specified file and
        # returns it. If the image cannot be read (because of missing
        # file, improper permissions, unsupported or invalid format),
        # the function returns an empty matrix
        # http://docs.opencv.org/3.0-beta/modules/imgcodecs/doc/reading_and_writing_images.html
        if grayscale:
            img_cv = cv2.imread(img, LOAD_GRAYSCALE)
        else:
            img_cv = cv2.imread(img, LOAD_COLOR)
        if img_cv is None:
            raise IOError("Failed to read %s because file is missing, "
                          "has improper permissions, or is an "
                          "unsupported or invalid format" % img)
    elif isinstance(img, numpy.ndarray):
        # don't try to convert an already-gray image to gray
        if grayscale and len(img.shape) == 3:  # and img.shape[2] == 3:
            img_cv = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        else:
            img_cv = img
    elif hasattr(img, 'convert'):
        # assume its a PIL.Image, convert to cv format
        img_array = numpy.array(img.convert('RGB'))
        img_cv = img_array[:, :, ::-1].copy()  # -1 does RGB -> BGR
        if grayscale:
            img_cv = cv2.cvtColor(img_cv, cv2.COLOR_BGR2GRAY)
    else:
        raise TypeError('expected an image filename, OpenCV numpy array, or PIL image')
    return img_cv


def locate_all_opencv(needle_image, haystack_image, grayscale=False, limit=10000, region=None, step=1,
                      confidence=0.999):
    """
    TODO - rewrite this
        faster but more memory-intensive than pure python
        step 2 skips every other row and column = ~3x faster but prone to miss;
            to compensate, the algorithm automatically reduces the confidence
            threshold by 5% (which helps but will not avoid all misses).
        limitations:
          - OpenCV 3.x & python 3.x not tested
          - RGBA images are treated as RBG (ignores alpha channel)
    """

    confidence = float(confidence)

    needle_image = _load_cv2(needle_image, grayscale)
    needle_height, needle_width = needle_image.shape[:2]
    haystack_image = _load_cv2(haystack_image, grayscale)

    if region:
        haystack_image = haystack_image[
                         region[1]:region[1] + region[3],
                         region[0]:region[0] + region[2]
                         ]
    else:
        region = (0, 0)  # full image; these values used in the yield statement
    if (haystack_image.shape[0] < needle_image.shape[0] or
            haystack_image.shape[1] < needle_image.shape[1]):
        # avoid semi-cryptic OpenCV error below if bad size
        raise ValueError('needle dimension(s) exceed the haystack image or region dimensions')

    if step == 2:
        confidence *= 0.95
        needle_image = needle_image[::step, ::step]
        haystack_image = haystack_image[::step, ::step]
    else:
        step = 1

    # get all matches at once, credit:
    # https://stackoverflow.com/questions/7670112/finding-a-subimage-inside-a-numpy-image/9253805#9253805
    result = cv2.matchTemplate(haystack_image, needle_image, cv2.TM_CCOEFF_NORMED)
    match_indices = numpy.arange(result.size)[(result > confidence).flatten()]
    matches = numpy.unravel_index(match_indices[:limit], result.shape)

    if len(matches[0]) == 0:
        return

    # use a generator for API consistency:
    matchx = matches[1] * step + region[0]  # vectorized
    matchy = matches[0] * step + region[1]

    # Order results before sending back
    ordered = sorted(zip(matchx, matchy), key=lambda p: result[p[1]][p[0]], reverse=True)
    for x, y in ordered:
        yield Box(x, y, needle_width, needle_height)



from typing import Dict

from selenium.webdriver import Ie  # noqa: F401, F403
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.ie.options import Options


def default_options(headless=False, download_folder_path=None, user_data_dir=None,
                    page_load_strategy="normal") -> Options:
    """Retrieve the default options for this browser curated by BotCity.

    Returns:
        Options: The Internet Explorer options.
    """
    ie_options = Options()
    ie_options.add_argument("-embedding")
    ie_options.add_argument("-extoff")
    ie_options.add_argument("-k")
    return ie_options


def default_capabilities() -> Dict:
    """Fetch the default capabilities for this browser.

    Returns:
        Dict: Dictionary with the default capabilities defined.
    """
    return DesiredCapabilities.INTERNETEXPLORER.copy()


def wait_for_downloads(driver):
    """Wait for all downloads to finish.
    """
    raise NotImplementedError('wait_for_downloads not yet supported')
